Attribute VB_Name = "Shots_Module"
Option Explicit

' shot structure definition
Public Type shot
 Exist As Boolean       ' exists?
 x As Long              ' x,y coords
 Y As Long
 Parent As Integer      ' who shooted it?
 ox As Long             ' old x,y coords
 oy As Long
 vx As Integer          ' x,y speed
 vy As Integer
 vr As Integer          ' speed module
 ind As Integer
 nrg As Single          ' energy carried
 range As Single        ' shot range
 travelled As Single    ' Distance travelled
 color As Long          ' colour
 shottype As Integer    ' carried location/value couple
 value As Integer       '
 lasttime As Integer    ' maximum time before desappearing
 pnode As node          ' node for sorting
 fromveg As Boolean     ' does shot come from veg?
 FromSpecie As String   ' Which species fired the shot
 Memloc As Integer      ' Memory location for custom poison and venom
 Memval As Integer      ' Value to insert into custom venom location
 DNA() As block         ' Somewhere to store genetic code for a virus
 contents As MaterialPacket 'food shots return a variety of bits.
End Type

Public Shots() As shot  ' array of shots
Public maxshots As Integer
Const shotdecay As Integer = 40 'increase to have shots lose power slower
Const ShellEffectiveness As Integer = 20 'how strong each unit of shell is
Const SlimeEffectiveness As Integer = 20 'how strong each unit of slime is against viruses

'
'   S H O T S   M A N A G E M E N T
'

' calculates the half brightness of a colour
' for a vaguely shiny effect in particles burst
Private Function HBrite(ByVal col As Long) As Long
  Dim b As Integer, g As Integer, r As Integer
  b = Int(col / 65536)
  col = col - (b * 65536)
  g = Int(col / 256)
  r = col - (g * 256)
  b = b / 2
  g = g / 2
  r = r / 2
  HBrite = RGB(r, g, b)
End Function

' same, but doubles
Public Function DBrite(ByVal col As Long) As Long
  Dim b As Long, g As Long, r As Long
  b = Int(col / 65536)
  col = col - (b * 65536)
  g = Int(col / 256)
  r = col - (g * 256)
  b = b + (255 - b) / 2
  g = g + (255 - g) / 2
  r = r + (255 - r) / 2
  DBrite = RGB(r, g, b)
End Function

' creates a shot shooted by robot n, with couple location/value
Public Sub newshot(n As Integer, shottype As Integer, val As Long, rngmultiplier As Single)
  Dim a As Integer
  Dim vs As Integer
  Dim ran As Single
  Dim ShAngle As Single
  If IsArrayBounded(Shots) = False Then ReDim Shots(100)
  a = FirstSlot
  If a > UBound(Shots) Then ReDim Preserve Shots(UBound(Shots) + 1)
  Shots(a).Parent = n
  Shots(a).FromSpecie = rob(n).fname    'Which species fired the shot
  Shots(a).fromveg = rob(n).Veg 'does shot come from a veg or not?
  Shots(a).ox = Shots(a).x
  Shots(a).oy = Shots(a).Y
  Shots(a).x = (rob(n).x + (RobSize / 2) + (rob(n).body / factor))
  Shots(a).Y = (rob(n).Y + (RobSize / 2) + (rob(n).body / factor))
  vs = RobSize / 3
  If val > 32000 Then val = 32000
  ran = Random(-2, 2) / 20
  If rob(n).mem(backshot) = 0 Then
    ShAngle = rob(n).aim        'forward shots
  Else
    ShAngle = angnorm(rob(n).aim - pi) 'backward shots
    rob(n).mem(backshot) = 0
  End If
  
  If rob(n).mem(aimshoot) > 0 Then '0 is the same as .shoot without any aiming
    While rob(n).mem(aimshoot) > 1256
      rob(n).mem(aimshoot) = rob(n).mem(aimshoot) - 1256
    Wend
    ShAngle = (rob(n).aim + rob(n).mem(aimshoot) / 200) 'Mod (2 * pi)
    rob(n).mem(aimshoot) = 0
  End If

  Shots(a).vx = absx(ShAngle + ran, vs, 0, 0, 0) + rob(n).vx ' set shot speed x seems to not work well at high bot speeds
  Shots(a).vy = absy(ShAngle + ran, vs, 0, 0, 0) + rob(n).vy ' set shot speed y
  'vs = Sqr(Shots(a).vx ^ 2 + Shots(a).vy ^ 2)
  Shots(a).vr = vs
  Shots(a).travelled = 0
  Shots(a).color = rob(n).color
  If shottype = -2 Then Shots(a).color = vbWhite
  Shots(a).Exist = True
  'shots(a).nrg = RobSize * 3                   'old shot range formula
  If rob(n).vbody > 10 Then
    Shots(a).range = Log(Abs(rob(n).vbody)) * 60 * rngmultiplier   'new range formula based on body size. Needs about 200 body to equal old method range. 32000 body doubles old range
    Shots(a).nrg = Shots(a).range
  Else
    Shots(a).range = RobSize * 3 * rngmultiplier
    Shots(a).nrg = RobSize * 3 * rngmultiplier
  End If
  If shottype = -7 Then
    Shots(a).nrg = rob(n).mem(338) * 20
    Shots(a).range = Shots(a).nrg
    Shots(a).color = vbCyan
  End If
  Shots(a).value = val
  Shots(a).shottype = shottype
  Shots(a).lasttime = ShotsLastTime
  Shots(a).Memloc = rob(n).mem(835)     'location for poison or venom to target
  If Shots(a).Memloc < 0 Then Shots(a).Memloc = 0
  If Shots(a).Memloc > 1000 Then Shots(a).Memloc = 1000
  Shots(a).Memval = rob(n).mem(836)     'value to insert into venom target location
  
  Set Shots(a).pnode = rlist.head
  sortshot a, Shots(a).vx
  If shottype = -7 Then copygene a, val
End Sub

' creates a generic particle with arbitrary x & y, vx & vy, etc
Public Sub createshot(x As Long, Y As Long, vx As Integer, vy As Integer, loc As Integer, par As Integer, val As Long, range As Single, col As Long)
  Dim a As Integer
  If IsArrayBounded(Shots) = False Then ReDim Shots(100)
  a = FirstSlot
  If a > UBound(Shots) Then ReDim Preserve Shots(UBound(Shots) + 1)
  Shots(a).Parent = par
  Shots(a).FromSpecie = rob(par).fname
  Shots(a).fromveg = rob(par).Veg
  Shots(a).x = x
  Shots(a).Y = Y
  Shots(a).vx = vx
  Shots(a).vy = vy
  Shots(a).vr = Sqr(vx ^ 2 + vy ^ 2)
  Shots(a).travelled = 0
  Shots(a).color = col
  Shots(a).Exist = True
  Shots(a).nrg = RobSize * 3
  Shots(a).nrg = range
  Shots(a).range = range
  Shots(a).value = val
  Shots(a).shottype = loc
  Shots(a).lasttime = ShotsLastTime
  Shots(a).Memloc = rob(par).mem(834)
  If Shots(a).Memloc < 0 Then Shots(a).Memloc = 0
  If Shots(a).Memloc > 1000 Then Shots(a).Memloc = 1000
  
  Set Shots(a).pnode = rlist.head
  sortshot a, Shots(a).vx
End Sub

' searches some place to insert the new shot in the
' shots array. Should be changed with a linked list!
Private Function FirstSlot() As Integer
  Dim t As Integer
  t = 1
  While Shots(t).Exist
    t = t + 1
    If t > UBound(Shots) Then GoTo exitloop
  Wend
exitloop:
  FirstSlot = t
  If FirstSlot > maxshots Then maxshots = FirstSlot
End Function

' keeps a double linked list of  shots sorted, for
' speeding up collision detection
Private Sub sortshot(n As Integer, vx As Integer)
  Dim nd As node
  With Shots(n)
    If SimOpts.Toroidal Then
      'wrap around torroidally
      If .x < 0 Then .x = .x + SimOpts.FieldWidth
      If .x > SimOpts.FieldWidth Then .x = .x - SimOpts.FieldWidth
     Else
      If .x < 0 Then .x = 0: .vx = -.vx
      If .x > SimOpts.FieldWidth Then .x = SimOpts.FieldWidth: .vx = -.vx
    End If
    Set nd = .pnode
    While .x > nd.xpos
      Set nd = nd.orn
    Wend
    While .x <= nd.xpos
      Set nd = nd.orp
    Wend
    Set .pnode = nd
  End With
End Sub

' calculates next shots position
Public Sub updateshots()
  Dim t As Integer
  Dim h As Integer
  Dim dx As Integer
  Dim sx As Integer
  Dim rp As Integer
  Dim jj As Integer
  Dim ti As Single
  Dim x As Long
  Dim Y As Long
  Dim v As Integer
  Dim onrg As Long
  For t = 1 To maxshots
    With Shots(t)
      If .Exist Then
        .lasttime = .lasttime - 1
        .ox = .x
        .oy = .Y
        x = .x
        Y = .Y
        v = .value
        .x = .x + .vx
        .Y = .Y + .vy
        If SimOpts.Pondmode Or Not SimOpts.Toroidal Then
          If .Y < 0 Then .Y = 0: .vy = -.vy
          If .Y > SimOpts.FieldHeight Then .Y = SimOpts.FieldHeight: .vy = -.vy
        End If
        .travelled = .travelled + .vr
        '.nrg = .nrg - (RobSize / 3) 'so that bots dueling against each other in flight don't have to worry about absolute veloctiy, only relative velocity
        'nrg loses next to nothing until right at the very end, where it decays rapidly.
        onrg = .nrg
        .nrg = .range * -Atn(.travelled / .range * shotdecay - shotdecay) / 1.55413
        If .nrg < 0 Then .nrg = 0
        If onrg - .nrg > 0 Then
          If x > 0 And Y > 0 Then
            Select Case .shottype   'transfers attributes from a dead shot tto the e-grid
              Case 0                'Poffs
                'envir.ValueToGrid x, Y, 7, 1  'add 1 to silica
                'envir.ValueToGrid x, Y, 8, 1  'add 1 to Calium
              Case -4
                'envir.ValueToGrid x, Y, 1, (onrg - .nrg) / .range * v 'add shot value to waste
              Case -2
                'envir.ValueToGrid x, Y, 1, (onrg - .nrg) / .range * v 'add shot value to waste
              Case -1
                'envir.ValueToGrid x, Y, 9, 1  'add 1 to temperature
            End Select
          End If
        End If
        If .nrg <= 0 Then .Exist = False
        sortshot t, .vx
        h = shotcoll(t)
        If h > 0 Then
          .Exist = False
          
          If .Parent = rob(h).Parent And rob(h).age < 10 Then GoTo exitif
          'babies born into a stream of shots from its parent shouldn't die
          'from those shots.  I can't imagine this temporary imunity can be
          'exploited, so it should be safe
          
          If .shottype > 0 Then
            If .shottype <> 340 Then
              Dim power As Long
              If .shottype = 312 Or .shottype = 313 Or .shottype = 824 Or .shottype = 826 And .value > 100 Then .value = 100
              
              If .nrg / 2 > rob(h).poison Then
                rob(h).mem(.shottype) = .value
              Else
                createshot Shots(t).x, Shots(t).Y, -.vx, -.vy, -5, h, .nrg / 2, Shots(t).range, vbYellow
                rob(h).poison = rob(h).poison - (.nrg / 2) * 0.9
                rob(h).Waste = rob(h).Waste + (.nrg / 2) * 0.1
                If rob(h).poison < 0 Then rob(h).poison = 0
                rob(h).mem(poison) = rob(h).poison
              End If
            End If
          Else
            If .shottype = -1 And Not rob(h).Wall Then releasenrg h, t
            If .shottype = -2 Then takenrg h, t
            If .shottype = -3 And h <> .Parent And Not rob(h).Wall Then takeven h, t
            If .shottype = -4 And Not rob(h).Wall Then takewaste h, t
            If .shottype = -5 And h <> .Parent And Not rob(h).Wall Then takepoison h, t
            If .shottype = -6 And Not rob(h).Wall Then releasebod h, t
            If .shottype = -7 And Not rob(h).Wall Then addgene h, t
          End If
          taste h, .ox, .oy, .shottype
        End If
exitif:
        If .lasttime < 0 Then .Exist = False
      End If
    End With
  Next t
End Sub

Public Sub Decay(n As Integer) 'corpse decaying as waste shot, energy shot or no shot
  Dim sh As Integer
  Dim va As Long
  rob(n).DecayTimer = rob(n).DecayTimer + 1
  If rob(n).DecayTimer >= SimOpts.Decaydelay Then
    rob(n).DecayTimer = 0
    If SimOpts.DecayType = 2 Then
      rob(n).aim = Rnd * 2 * pi
      rob(n).aimx = Cos(rob(n).aim)
      rob(n).aimy = Sin(rob(n).aim)
      sh = -4
      va = SimOpts.Decay
      newshot n, sh, va, 1
    End If
    If SimOpts.DecayType = 3 Then
      rob(n).aim = Rnd * 2 * pi
      rob(n).aimx = Cos(rob(n).aim)
      rob(n).aimy = Sin(rob(n).aim)
      sh = -2
      va = SimOpts.Decay
      newshot n, sh, va, 1
    End If
    rob(n).body = rob(n).body - SimOpts.Decay / 10
    rob(n).Radius = rob(n).body / factor
  End If
End Sub
Public Sub defacate(n As Integer) 'only used to get rid of massive amounts of waste
  Dim sh As Integer
  Dim va As Long
  sh = -4
  va = 200
  If va > rob(n).Waste Then va = rob(n).Waste
  If rob(n).Waste > 32000 Then rob(n).Waste = 31500: va = 500
  rob(n).Waste = rob(n).Waste - va
  rob(n).nrg = rob(n).nrg - SimOpts.CostChem / (rob(n).Numties + 1)
  newshot n, sh, va, 1
  rob(n).Pwaste = rob(n).Pwaste + va / 1000
End Sub

' robot n, hit by shot t, releases energy
Public Sub releasenrg(n As Integer, t As Integer)
  'n=robot number
  't=shot number
  Dim vx As Integer
  Dim vy As Integer
  Dim vs As Integer
  Dim vr As Single
  Dim power As Long
  Dim h2 As Integer
  Dim range As Single
  Dim scalingfactor As Single
  Dim Newangle As Single
      
  h2 = Half / 4
  vx = (Shots(t).vx - rob(n).vx) / -2 + rob(n).vx
  vy = (Shots(t).vy - rob(n).vy) / -2 + rob(n).vy
  
  If SimOpts.EnergyExType Then
    power = Shots(t).value * (Shots(t).nrg / Shots(t).range)
  Else
    power = SimOpts.EnergyFix
  End If
  
  If power > rob(n).nrg And Not rob(n).Corpse Then power = rob(n).nrg
  If rob(n).Corpse Then power = power * 0.5
  
  range = Shots(t).range     'new range formula based on range of incoming shot
  
  If rob(n).poison > power Then 'create poison shot
    createshot Shots(t).x, Shots(t).Y, vx, vy, -5, n, power, range, vbYellow
    rob(n).Waste = rob(n).Waste + (power * 0.1)
    rob(n).poison = rob(n).poison - (power * 0.9)
    If rob(n).poison < 0 Then rob(n).poison = 0
    rob(n).mem(poison) = rob(n).poison
  Else ' create energy shot
    createshot Shots(t).x, Shots(t).Y, vx, vy, -2, n, power, range, vbWhite
    rob(n).nrg = rob(n).nrg - power * 0.9  'some of shot comes from nrg
    rob(n).body = rob(n).body - power * 0.01 'some of shot comes from body
    rob(n).Radius = rob(n).body / factor
  End If
  If rob(n).body < 0 Or rob(n).nrg < 0 Then
    rob(Shots(t).Parent).kills = rob(Shots(t).Parent).kills + 1
    rob(Shots(t).Parent).mem(220) = rob(Shots(t).Parent).kills
  End If
  Shots(t).Exist = False
End Sub
Private Sub releasebod(n As Integer, t As Integer) 'a robot is shot by a -6 shot and releases energy directly from body points
  'much more effective against a corpse
  Dim vx As Integer
  Dim vy As Integer
  Dim range As Single
  Dim power As Long
  Dim Shell As Long
  Dim h2 As Integer
   
  h2 = Half / 4
  vx = (Shots(t).vx - rob(n).vx) / -2 + rob(n).vx
  vy = (Shots(t).vy - rob(n).vy) / -2 + rob(n).vy
  
  If SimOpts.EnergyExType Then
    power = Shots(t).nrg / Shots(t).range * Shots(t).value
  Else
    power = SimOpts.EnergyFix
  End If
  
  Shell = rob(n).Shell * ShellEffectiveness
  
  If power < Shell Then
    rob(n).Shell = rob(n).Shell - power / ShellEffectiveness
    If rob(n).Shell < 0 Then rob(n).Shell = 0
    Exit Sub
  Else
    Dim temp As Long
    temp = rob(n).Shell
    rob(n).Shell = rob(n).Shell - power / ShellEffectiveness
    power = power - temp * ShellEffectiveness
    If rob(n).Shell < 0 Then rob(n).Shell = 0
  End If
  
  If power < 0.1 Then Exit Sub
  
  range = Shots(t).range      'new range formula based on range of incoming shot
  
  ' create energy shot
  If rob(n).Corpse = True Then
    power = power * 4
    If power > rob(n).body * 10 Then power = rob(n).body * 10
    rob(n).body = rob(n).body - power / 10      'all energy comes from body
    rob(n).Radius = rob(n).body / factor
  Else
    Dim leftover As Long
    
    rob(n).nrg = rob(n).nrg - power * 0.2       'some of shot comes from nrg
    rob(n).body = rob(n).body - (power * 0.08)  'some of shot comes from body
        
    leftover = 0
    If rob(n).nrg < 0 Then leftover = leftover + (-rob(n).nrg): rob(n).nrg = 0
    If rob(n).body < 0 Then leftover = leftover + (-rob(n).body * 10): rob(n).body = 0
    
    With rob(n)
    If leftover > 0 Then
      If .nrg > 0 And .nrg > leftover Then
        .nrg = .nrg - leftover
        leftover = 0
      ElseIf .nrg > 0 And .nrg < leftover Then
        leftover = leftover - .nrg
        .nrg = 0
      End If
      
      If .body > 0 And .body * 10 > leftover Then
        .body = .body - leftover
        leftover = 0
      ElseIf rob(n).body > 0 And rob(n).body * 10 < leftover Then
        leftover = leftover - .body
        .body = 0
      End If
      
      If leftover > 0 Then power = power - leftover
    End If
    End With
    rob(n).Radius = rob(n).body / factor
  End If
  If rob(n).body < 0 Or rob(n).nrg < 0 Then
    rob(n).Dead = True
    rob(Shots(t).Parent).kills = rob(Shots(t).Parent).kills + 1
    rob(Shots(t).Parent).mem(220) = rob(Shots(t).Parent).kills
  End If
  createshot Shots(t).x, Shots(t).Y, vx, vy, -2, n, power, range, vbWhite
  Shots(t).Exist = False
End Sub

' robot n takes the energy carried by shot t
Private Sub takenrg(n As Integer, t As Integer)
  Dim partial As Long
  
  partial = Shots(t).nrg / Shots(t).range * Shots(t).value
  
  If rob(n).Corpse Then Exit Sub
  rob(n).nrg = rob(n).nrg + partial * 0.8       'most of energy goes to nrg
  rob(n).body = rob(n).body + (partial * 0.019) 'a bit goes to body
  rob(n).Waste = rob(n).Waste + partial * 0.001  'tiny amount goes to waste
  If rob(n).nrg > 32000 Then rob(n).nrg = 32000
  If rob(n).body > 32000 Then rob(n).body = 32000
  Shots(t).Exist = False
  rob(n).Radius = rob(n).body / factor
End Sub
'  robot takes a venomous shot and becomes seriously messed up
Private Sub takeven(n As Integer, t As Integer)
  Dim power As Long
  
  power = Shots(t).nrg / Shots(t).range * Shots(t).value
  
  If Shots(t).Memloc = 340 Or power < 1 Then Exit Sub
  
  If Shots(t).FromSpecie = rob(n).fname Then   'Robot is imune to venom from his own species
    rob(n).venom = rob(n).venom + power   'Robot absorbs venom fired by conspec
    rob(n).mem(825) = rob(n).venom
  Else
    If power < rob(n).Shell * ShellEffectiveness Then
      rob(n).Shell = rob(n).Shell - power * ShellEffectiveness
      If rob(n).Shell < 0 Then rob(n).Shell = 0
    Else
      Dim temp As Long
      temp = power
      power = power - rob(n).Shell * ShellEffectiveness
      rob(n).Shell = rob(n).Shell - temp / ShellEffectiveness
      If rob(n).Shell < 0 Then rob(n).Shell = 0
    End If
    
    If power < 1 Then Exit Sub
    
    rob(n).Paralyzed = True                 'OK not actually paralyzed with the new Memloc/Memval couplet
    rob(n).Paracount = rob(n).Paracount + power
    If Shots(t).Memloc > 0 Then
      If Shots(t).Memloc > 999 Then Shots(t).Memloc = 999
      rob(n).Vloc = Shots(t).Memloc
    Else
      rob(n).Vloc = Int(Rnd * 1000) + 1
    End If
    rob(n).Vval = Shots(t).Memval
  End If
  Shots(t).Exist = False
End Sub
'  Robot n takes shot t and adds its value to his waste reservoir
Private Sub takewaste(n As Integer, t As Integer)
  Dim power As Long
  
  power = Shots(t).nrg / Shots(t).range * Shots(t).value
  
  rob(n).Waste = rob(n).Waste + power
  Shots(t).Exist = False
End Sub
' Robot receives poison shot and becomes disorientated
Private Sub takepoison(n As Integer, t As Integer)
  If Shots(t).FromSpecie = rob(n).fname Then    'Robot is imune to poison from his own species
    rob(n).poison = rob(n).poison + Shots(t).value 'Robot absorbs poison fired by conspecs
    rob(n).mem(827) = rob(n).poison
  Else
    Dim power As Long
    
    power = Shots(t).nrg / Shots(t).range * Shots(t).value
    
    rob(n).Poisoned = True
    rob(n).Poisoncount = rob(n).Poisoncount + power
    If rob(n).Poisoncount > 32000 Then rob(n).Poisoncount = 32000
    If Shots(t).Memloc > 0 Then
      rob(n).Ploc = Shots(t).Memloc
    Else
      rob(n).Ploc = Int(Rnd * 1000) + 1
    End If
  End If
  Shots(t).Exist = False
End Sub

' checks the collisions between robots and shots
Private Function shotcoll(n As Integer) As Integer
  Dim nd As node
  Dim x As Long
  Dim Y As Long
  Dim fxn As Single
  x = Shots(n).x
  Y = Shots(n).Y
  
  If x > SimOpts.FieldWidth Or x < 0 Or Y > SimOpts.FieldHeight Or Y < 0 Then
    If SimOpts.Toroidal = True Then
      If x > SimOpts.FieldWidth Then x = x - SimOpts.FieldWidth
      If x < 0 Then x = x + SimOpts.FieldWidth
      If Y > SimOpts.FieldHeight Then Y = Y - SimOpts.FieldHeight
      If Y < 0 Then Y = Y + SimOpts.FieldHeight
    ElseIf SimOpts.Pondmode = True Then
      If x > SimOpts.FieldWidth Then x = x - SimOpts.FieldWidth
      If x < 0 Then x = x + SimOpts.FieldWidth
    Else 'wrap the shot around
      If x > SimOpts.FieldWidth Then x = x - SimOpts.FieldWidth
      If Y > SimOpts.FieldHeight Then Y = Y - SimOpts.FieldHeight
      If x < 0 Then x = x + SimOpts.FieldWidth
      If Y < 0 Then Y = Y + SimOpts.FieldHeight
    End If
    Shots(n).x = x
    Shots(n).Y = Y
  Else
    Set nd = Shots(n).pnode
    While x - nd.xpos < RobSize
      fxn = Y - rob(nd.robn).Y - Half
      If fxn < 0 Then fxn = -fxn
      If fxn <= Half + rob(nd.robn).body / factor Then
        If Shots(n).Parent <> nd.robn Then shotcoll = nd.robn
      End If
      Set nd = nd.orp
    Wend
  End If
End Function

Public Sub Vshoot(n As Integer)
'here we shoot a virus
  Dim sh As Integer
  Dim va As Long
  sh = -7
  va = rob(n).mem(335)
  If va <> 0 Then
    rob(n).nrg = rob(n).nrg - va * (SimOpts.CostChem / (rob(n).Numties + 1))
    If rob(n).Multibot Then
      va = va * (rob(n).Numties + 1) 'make virus stronger when fired by MB
    End If
    newshot n, sh, va, 1
  End If
  rob(n).mem(338) = 0
End Sub

' copy gene number p from robot that fired shot n into shot n dna (virus)
Public Function copygene(n As Integer, p As Long)
  Dim k As Integer, t As Integer
  Dim par As Integer
  Dim genelen As Integer
  Dim i As Integer
  Dim f As Integer
  Dim q As Integer
  
  par = Shots(n).Parent
  k = CountGenes(rob(par).DNA)
  If p > k Then p = k
  If p < 1 Then p = 1
  q = p
  ' target gene is not beyond end of DNA
  i = GenePos(rob(par).DNA, q)
  f = NextStop(rob(par).DNA, i)
  genelen = f - i
  
  If genelen < 0 Then
    Shots(n).Exist = False
    Exit Function
  End If
  
  ReDim Preserve Shots(n).DNA(genelen + 1)
  For t = 1 To genelen + 1
    Shots(n).DNA(t) = rob(par).DNA(t + i - 1)
  Next t
  k = p
End Function
' adds gene from shot p to robot n dna
Public Function addgene(n As Integer, p As Integer) As Integer
  Dim k As Integer, t As Integer
  Dim i As Integer
  Dim f As Integer
  Dim dist As Integer
  Dim fname As String
  Dim Insert As Integer
  Dim vlen As Integer   'length of the DNA code of the virus
  Dim Position As Integer   'gene position to insert the virus
  Dim power As Long
  
  power = Shots(p).nrg / Shots(p).range * Shots(p).value
  
  If power < rob(n).Slime * SlimeEffectiveness Then
    rob(n).Slime = rob(n).Slime - power / SlimeEffectiveness
    Exit Function
  Else
    Dim temp As Long
    temp = rob(n).Slime
    
    rob(n).Slime = rob(n).Slime - power / SlimeEffectiveness
    power = power - rob(n).Slime * SlimeEffectiveness
    If rob(n).Slime < 1 Then rob(n).Slime = 1
  End If
  
  k = CountGenes(rob(n).DNA)
  Position = Int(Rnd() * (k - 1) + 1)   'randomize the gene number
  If Position = 0 Then Exit Function    'shouldn't be possible anyway
  i = GenePos(rob(n).DNA, Position)     'DNA position for start of this gene (cond)
  f = NextStop(rob(n).DNA, i)           'DNA position of end of gene (stop)
  vlen = VirusLen(Shots(p).DNA)         'length of the virus DNA
  Insert = f + 1                        'Position to insert the virus
  If f < 0 Then Exit Function           'sometimes f comes back as -1 ????
  fname = rob(n).fname                  'just to help debug
  
  If MakeSpace(rob(n).DNA, Insert, vlen) Then   'Moves gene back to make space
    dist = Insert - i                   'length of the gene being moved. This doesn't add up right??
    For t = 1 To vlen
      rob(n).DNA(t + f) = Shots(p).DNA(t)
    Next t
  End If
  makecondlist n
  makeoccurrlist n
  countconditions n
  rob(n).DnaLen = DnaLen(rob(n).DNA)
  rob(n).genenum = CountGenes(rob(n).DNA)
End Function

Private Function VirusLen(ByRef DNA() As block) As Integer
Dim k As Integer
  While Not (DNA(k).tipo = 4 And DNA(k).value = 3)
    k = k + 1
    If k > UBound(DNA) Then
      k = k - 1
      DNA(k).tipo = 4 And DNA(k).value = 3
    End If
  Wend
exitme:
  If Not (DNA(k).tipo = 4 And DNA(k).value = 3) Then k = -1
  VirusLen = k
End Function

Private Function IsArrayBounded(ByRef ArrayIn() As shot) As Boolean
On Error Resume Next
IsArrayBounded = (UBound(ArrayIn) >= LBound(ArrayIn))
End Function
